package edu.qrg.vmodel;

import edu.qrg.utils.*;
import edu.qrg.utils.CMLio.CMLable;
import edu.qrg.vmodel.Definition;
import edu.qrg.vmodel.Instance;
import edu.qrg.utils.KeyValuePair;
import edu.qrg.utils.KeyValuePair.KVPinterface;
import edu.qrg.vmodel.Ontology.Element;
import edu.qrg.utils.PList.PLinterface;
import edu.qrg.utils.XMLio.XMLable;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.lang.ClassNotFoundException;
import java.lang.Cloneable;
import java.lang.Object;
import java.lang.String;
import java.util.Date;
import java.util.Enumeration;
import java.util.Vector;
import java.util.*;
import javax.swing.event.TreeModelListener;
import javax.swing.event.TreeModelEvent;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreeNode;
import javax.swing.tree.TreePath;

/** <PRE> Sun Java 1.3.1 -Sourcepath QRG/VModel/development/code/ontology
* ----------------------------------------------------------------------------
*  2003 Northwestern University Qualitative Reasoning Group Kenneth D. Forbus
* ----------------------------------------------------------------------------
* File name: Model.java
*    System: VModel
*     Since: v0.6
*    Author: Leo C. Ureel II
* ----------------------------------------------------------------------------
*   Class: Model
* Purpose: Defines properties and methods common to all Models.
* ----------------------------------------------------------------------------
* </PRE> */
public class Model extends Definition
		implements 	Cloneable, Serializable, VModelConstants, Ontology.Category {
	// Constructors
	// ------------------------------------------------------------------------
	/** <PRE>
	* ------------------------------------------------------------------------
	* Purpose: General Initialization.
	* ------------------------------------------------------------------------
	* </PRE>
	*/
	public Model () {
		super();
    } // default constructor

	// Ontology.Category methods
	// -----------------------------------------------------------------------
	/** Purpose: Returns the string filename of an image representing
	             the definition as a tree node in an Open or Close state. */
	public String icon (boolean expanded) {
		String result = icon();
		if (expanded) result += "Open";
		else result += "Closed";
		return(result);
	} // method

	// Properties & Accessors
	// ------------------------------------------------------------------------
	/** Use accessor.
	    @see #localOntology()
	    @serial */
	protected Ontology localOntology = null;
	/** <PRE>
	* ------------------------------------------------------------------------
	* Purpose: Local stash of modeling definitions.
	* ------------------------------------------------------------------------
	* </PRE>
	*/
	public Ontology localOntology () {
		if (localOntology == null) {
			localOntology = new Ontology();
			localOntology.name(name());
		} // end if
		return(localOntology);
	} // reader

	/** Use accessor.
	    @see #instances()
	    @serial */
	protected Vector instances = null;
	/** <PRE>
	* ------------------------------------------------------------------------
	* Purpose: Instanciations of definitions comprising the model.
	* ------------------------------------------------------------------------
	* </PRE>
	*/
	public Vector instances () {
		if (instances == null) instances = new Vector();
		return(instances);
	} // reader

	/** Use accessor.
	    @see #log()
	    @serial */
	protected Vector log = null;
	/** <PRE>
	* ------------------------------------------------------------------------
	* Purpose: Log of modeling actions.
	* ------------------------------------------------------------------------
	* </PRE>
	*/
	public Vector log () {
		if (log == null) log = new Vector();
		return(log);
	} // reader

	// Methods
	// -----------------------------------------------------------------------
	/** <PRE>
	* ------------------------------------------------------------------------
	* Purpose: Find the definition with the specified id in the local library,
	*          if one exists there.
	* ------------------------------------------------------------------------
	* </PRE>
	*/
	public Definition findLocal (String id) {
		Definition result = (Definition) localOntology().findElement(id);
		return(result);
	} // method

	/** <PRE>
	* ------------------------------------------------------------------------
	* Purpose: Find the instance with the specified id in the model,
	*          if one exists there.
	* ------------------------------------------------------------------------
	* </PRE>
	*/
	public Instance findInstance (String id) {
		Instance result = null;
		Instance mumble = null;
		if (! ((id == null) || (id.equalsIgnoreCase("")))) {
			for (Enumeration o = instances().elements() ; o.hasMoreElements() ;) {
				mumble =  (Instance) o.nextElement();
				if (id.equalsIgnoreCase(mumble.id())) {
					result = mumble;
					break;
				} // end if
			} // end for
		} // end if
//		if ((result == null) && (!id.equals(""))) {
//			QRGutils.debug("Model.findInstance", "ID not found. id=" + id + " instances=" + instances());
//		} // end if
		return(result);
	} // method

	// Log
	// -----------------------------------------------------------------------
	/** Store information in a model log. */
	public Vector log (String text) {
		Date currentTime = new Date();
		String datetime = QRGutils.dateformat().format(currentTime);
		log.addElement(datetime + " " + VModelPreferences.self().getString("USERID") + " " + text);
		return(log);
	} // method

	/** Prefix log information with a code and add to model log. */
	public Vector log (String code, String text) {
		Date currentTime = new Date();
		String datetime = QRGutils.dateformat().format(currentTime);
		log().addElement(datetime + " " + VModelPreferences.self().getString("USERID") + " " + code + " : " + text);
		return(log);
	} // method

	// File I/O
	// ------------------------------------------------------------------------
	/** <PRE>
	* ------------------------------------------------------------------------
	* Purpose: Save the model to the specified file.
	* Notes:
	* * Currently saves as java serialized file.
	* * Also saves in XML and CML formats.
	* ------------------------------------------------------------------------
	* </PRE>
	*/
	public void save (String filename) {
		try {
			FileOutputStream ostream = new FileOutputStream(filename);
			ObjectOutputStream p = new ObjectOutputStream(ostream);
			log("save", "filepath= " + filename);
			p.writeObject(this);
			p.flush();
			ostream.close();
			XMLio.save(QRGutils.replaceExtension(filename, XML_SUFFIX), this);
			CMLio.save(QRGutils.replaceExtension(filename, CML_SUFFIX), this);
		//  What about ostream? Potential leak in catches.
		} catch (FileNotFoundException e) {
			QRGutils.debug("Model.save", " filename=" + filename + " Error: \n" + e);
			QRGutils.debugStackTrace(e);
		} catch (IOException e1) {
			QRGutils.debug("Model.save", " filename=" + filename + " Error: \n" + e1);
			QRGutils.debugStackTrace(e1);
		} // try catch
	} // method save

	/** <PRE>
	* ------------------------------------------------------------------------
	* Purpose: Loads a serialized model.
	* ------------------------------------------------------------------------
	* </PRE>
	*/
	public static Definition load (String filepath) {
		Model result = null;
		try {
			FileInputStream istream = new FileInputStream(filepath);
			ObjectInputStream p = new ObjectInputStream(istream);
			result = (Model) p.readObject();
			if (result != null) {
				result.log("load", "filepath= " + filepath);
			} // end if
			istream.close();
		//  What about istream? Potential leak in catches.
		} catch (FileNotFoundException e) {
			QRGutils.debug("Model.load", " Error: \n" + e);
			QRGutils.debug("Model.load", " Error filepath=" + filepath);
		} catch (IOException e) {
			QRGutils.debug("Model.load", " Error: \n" + e);
			QRGutils.debug("Model.load", " Error filepath=" + filepath);
		} catch (ClassNotFoundException e) {
			QRGutils.debug("Model.load", " Error: \n" + e);
			QRGutils.debug("Model.load", " Error filepath=" + filepath);
		} // try catch
		return(result);
	} // method

	// CMLio.CMLable
	// ----------------------------------------------------------------------
	/** <PRE>
	* ------------------------------------------------------------------------
	* Purpose: Returns the CML command used to represent the object.
	* ------------------------------------------------------------------------
	* </PRE>
	*/
	public String CMLcommand () { return("defmodel"); }

	/** <PRE>
	* ------------------------------------------------------------------------
	* Purpose: Stringifies the object into CML format.
	* ------------------------------------------------------------------------
	* </PRE>
	*/
	public String toCML () {
		String result = "";
		CMLio.CMLable element = null;
		// Should we import and print definitions referenced from libraries?
//		// 1. Print local library definitions
//		for (Enumeration enum = localOntology().children(); enum.hasMoreElements();) {
//			element = (CMLio.CMLable) enum.nextElement();
//			result += element.toCML() + QRGutils.eol();
//		} // end for
		// 2. Print model definition.
		result += super.toCML() + QRGutils.eol() + QRGutils.eol();
		// 3. Print Instance
//		result += ";; Instances" + QRGutils.eol();
//		result += ";; ---------------------------------------------------------------------------" + QRGutils.eol();
//		for (Enumeration enum = instances().elements(); enum.hasMoreElements();) {
//			element = (CMLio.CMLable) enum.nextElement();
//			result += element.toCML() + QRGutils.eol();
//		} // end for
		// 4. Print log
		result += ";; Event Log" + QRGutils.eol();
		result += ";; ---------------------------------------------------------------------------" + QRGutils.eol();
		for (Enumeration e = log().elements() ; e.hasMoreElements() ;) {
			result += (";; " + (String) e.nextElement() + QRGutils.eol());
		} // end for
		return(result);
	} // method

	// XMLio.XMLable
	// ----------------------------------------------------------------------
	/** <PRE>
	* ------------------------------------------------------------------------
	* Purpose: Stringify the object into the XML format.
	* ------------------------------------------------------------------------
	* </PRE>
	*/
	public String xmlify () {
		String result = "";
		// 1. print model definition
		result += super.xmlify() + QRGutils.eol() + QRGutils.eol();
		// 2. print log.
		result += XMLio.commentate("Event Log") + QRGutils.eol();
		result += XMLio.commentate("---------------------------------------------------------------------------") + QRGutils.eol();
		for (Enumeration e = log().elements() ; e.hasMoreElements() ;) {
			result += (XMLio.commentate((String) e.nextElement()) + QRGutils.eol());
		} // end for
		return(result);
	} // method


	/** <PRE>
	* ------------------------------------------------------------------------
	* Purpose: Initializes slots after object is created by XML loader.
	* ------------------------------------------------------------------------
	* </PRE>
	*/
	public void XMLioObjectInit (Vector attr, Vector elmnt) {
		super.XMLioObjectInit(attr, elmnt);
		Object mumble;
		Instance fumble;
		Enumeration enum;
		for(enum = connectors().elements(); enum.hasMoreElements();) {
			mumble = enum.nextElement();
			if (mumble instanceof Instance) {
				fumble = (Instance) mumble;
				fumble.model(this);
				instances().addElement(fumble);
			}
		}
		for(enum = connections().elements(); enum.hasMoreElements();) {
			mumble = enum.nextElement();
			if (mumble instanceof Instance) {
				fumble = (Instance) mumble;
				fumble.model(this);
				instances().addElement(fumble);
			}
		}
	}


	// TreeModel Methods
	// -----------------------------------------------------------------------
	/** Use accessors. */
	protected transient Vector treeModelListeners = null;
	/** Stash for listeners. Default is new Vector(). */
	public Vector treeModelListeners () {
		if (treeModelListeners == null) treeModelListeners = new Vector();
		return(treeModelListeners);
	} // reader

	/** Add a tree model listener.  Wow! These are exciting comments! */
	public void addTreeModelListener(TreeModelListener l) {
		treeModelListeners().addElement(l);
	} // method

	/** Remove tree model listener. */
	public void removeTreeModelListener(TreeModelListener l) {
		treeModelListeners().removeElement(l);
	} // method

	/** Returns child tree node. */
	public Object getChild(Object category, int index) {
		return(((TreeNode) category).getChildAt(index));
	} // method

	/** Returns number of child nodes. */
	public int getChildCount(Object category) {
		return(((TreeNode) category).getChildCount());
	} // method

	/** Returns location of child node in tree. */
	public int getIndexOfChild(Object category, Object child) {
		return(((TreeNode) category).getIndex((TreeNode) child));
	} // method

	/** Return root node of tree. */
	public Object getRoot() { return(this); } // method

	/** Return true if node has no children. */
	public boolean isLeaf(Object node) {
		return(((TreeNode) node).isLeaf());
	} // method

	/** Does this ever get called?  How can we use it to allow renaming of
	    ontological elements in the tree display? */
	public void valueForPathChanged(TreePath path, Object newValue) {
		System.out.println("Ontology.valueForPathChanged path=" + path + " newValue=" + newValue);
	} // method

    /*
     * Notify all listeners that have registered interest for
     * notification on this event type.  The event instance
     * is lazily created using the parameters passed into
     * the fire method.
     * @see EventListenerList
     */
    protected void fireTreeNodesChanged(Object source, Object[] path,
                                        int[] childIndices,
                                        Object[] children) {
       	TreeModelEvent e = null;
       	TreeModelListener l = null;
		for (Enumeration enum = treeModelListeners().elements(); enum.hasMoreElements();) {
			l = (TreeModelListener) enum.nextElement();
                if (e == null)
                    e = new TreeModelEvent(source, path,
                                           childIndices, children);
                l.treeNodesChanged(e);
        }
    }

    /*
     * Notify all listeners that have registered interest for
     * notification on this event type.  The event instance
     * is lazily created using the parameters passed into
     * the fire method.
     * @see EventListenerList
     */
    protected void fireTreeNodesInserted(Object source, Object[] path,
                                        int[] childIndices,
                                        Object[] children) {
       	TreeModelEvent e = null;
       	TreeModelListener l = null;
		for (Enumeration enum = treeModelListeners().elements(); enum.hasMoreElements();) {
			l = (TreeModelListener) enum.nextElement();
                // Lazily create the event:
                if (e == null)
                    e = new TreeModelEvent(source, path,
                                           childIndices, children);
                l.treeNodesInserted(e);
        }
    }

    /*
     * Notify all listeners that have registered interest for
     * notification on this event type.  The event instance
     * is lazily created using the parameters passed into
     * the fire method.
     * @see EventListenerList
     */
    protected void fireTreeNodesRemoved(Object source, Object[] path,
                                        int[] childIndices,
                                        Object[] children) {
       	TreeModelEvent e = null;
       	TreeModelListener l = null;
		for (Enumeration enum = treeModelListeners().elements(); enum.hasMoreElements();) {
			l = (TreeModelListener) enum.nextElement();
                if (e == null)
                    e = new TreeModelEvent(source, path,
                                           childIndices, children);
                l.treeNodesRemoved(e);
        }
    }

    /*
     * Notify all listeners that have registered interest for
     * notification on this event type.  The event instance
     * is lazily created using the parameters passed into
     * the fire method.
     * @see EventListenerList
     */
    protected void fireTreeStructureChanged(Object source, Object[] path,
                                        int[] childIndices,
                                        Object[] children) {
       	TreeModelEvent e = null;
       	TreeModelListener l = null;
		for (Enumeration enum = treeModelListeners().elements(); enum.hasMoreElements();) {
			l = (TreeModelListener) enum.nextElement();
                if (e == null)
                    e = new TreeModelEvent(source, path,
                                           childIndices, children);
                l.treeStructureChanged(e);
        }
    }

    /**
      * Invoke this method after you've changed how node is to be
      * represented in the tree.
      */
 //   public void nodeChanged(TreeNode node) {
 //       if(listenerList != null && node != null) {
 //           TreeNode         parent = node.getParent();
//
//            if(parent != null) {
//                int        anIndex = parent.getIndex(node);
//                if(anIndex != -1) {
//                    int[]        cIndexs = new int[1];
//
//                    cIndexs[0] = anIndex;
//                   nodesChanged(parent, cIndexs);
//                }
//            }
//	    else if (node == getRoot()) {
//		nodesChanged(node, null);
//	    }
//        }
//    }

    /**
     * Invoke this method if you've modified the TreeNodes upon which this
     * model depends.  The model will notify all of its listeners that the
     * model has changed below the node <code>node</code> (PENDING).
     */
    public void reload(TreeNode node) {
        if(node != null) {
            fireTreeStructureChanged(this, getPathToRoot(node), null, null);
        }
    }

    /**
      * Invoke this method after you've inserted some TreeNodes into
      * node.  childIndices should be the index of the new elements and
      * must be sorted in ascending order.
      */
//    public void nodesWereInserted(TreeNode node, int[] childIndices) {
//        if(listenerList != null && node != null && childIndices != null
//           && childIndices.length > 0) {
//            int               cCount = childIndices.length;
//            Object[]          newChildren = new Object[cCount];
//
//            for(int counter = 0; counter < cCount; counter++)
//                newChildren[counter] = node.getChildAt(childIndices[counter]);
//            fireTreeNodesInserted(this, getPathToRoot(node), childIndices,
//                                  newChildren);
//        }
//    }

    /**
      * Invoke this method after you've removed some TreeNodes from
      * node.  childIndices should be the index of the removed elements and
      * must be sorted in ascending order. And removedChildren should be
      * the array of the children objects that were removed.
      */
    public void nodesWereRemoved(TreeNode node, int[] childIndices,
                                 Object[] removedChildren) {
        if(node != null && childIndices != null) {
            fireTreeNodesRemoved(this, getPathToRoot(node), childIndices,
                                 removedChildren);
        }
    }

    /**
      * Invoke this method after you've changed how the children identified by
      * childIndicies are to be represented in the tree.
      */
    public void nodesChanged(TreeNode node, int[] childIndices) {
        if(node != null) {
	    if (childIndices != null) {
		int            cCount = childIndices.length;

		if(cCount > 0) {
		    Object[]       cChildren = new Object[cCount];

		    for(int counter = 0; counter < cCount; counter++)
			cChildren[counter] = node.getChildAt
			    (childIndices[counter]);
		    fireTreeNodesChanged(this, getPathToRoot(node),
					 childIndices, cChildren);
		}
	    }
	    else if (node == getRoot()) {
		fireTreeNodesChanged(this, getPathToRoot(node), null, null);
	    }
        }
    }

    /**
      * Invoke this method if you've totally changed the children of
      * node and its childrens children...  This will post a
      * treeStructureChanged event.
      */
    public void nodeStructureChanged(TreeNode node) {
        if(node != null) {
           fireTreeStructureChanged(this, getPathToRoot(node), null, null);
        }
    }

    /**
     * Builds the parents of node up to and including the root node,
     * where the original node is the last element in the returned array.
     * The length of the returned array gives the node's depth in the
     * tree.
     *
     * @param aNode the TreeNode to get the path for
     * @param an array of TreeNodes giving the path from the root to the
     *        specified node.
     */
    public TreeNode[] getPathToRoot(TreeNode aNode) {
        return getPathToRoot(aNode, 0);
    }

    /**
     * Builds the parents of node up to and including the root node,
     * where the original node is the last element in the returned array.
     * The length of the returned array gives the node's depth in the
     * tree.
     *
     * @param aNode  the TreeNode to get the path for
     * @param depth  an int giving the number of steps already taken towards
     *        the root (on recursive calls), used to size the returned array
     * @return an array of TreeNodes giving the path from the root to the
     *         specified node
     */
    protected TreeNode[] getPathToRoot(TreeNode aNode, int depth) {
        TreeNode[]              retNodes;
	// This method recurses, traversing towards the root in order
	// size the array. On the way back, it fills in the nodes,
	// starting from the root and working back to the original node.

        /* Check for null, in case someone passed in a null node, or
           they passed in an element that isn't rooted at root. */
        if(aNode == null) {
            if(depth == 0)
                return null;
            else
                retNodes = new TreeNode[depth];
        }
        else {
            depth++;
            if(aNode == getRoot())
                retNodes = new TreeNode[depth];
            else
                retNodes = getPathToRoot(aNode.getParent(), depth);
            retNodes[retNodes.length - depth] = aNode;
        }
        return retNodes;
    }

    // Methods from Ontology.Element
    // -----------------------------------------------------------------------
	/** Finds the ontological element with the specified id.
	*   Otherwise returns null.
	* * Recurses into categories.
	*/
	public Ontology.Element findElement (String id) {
		Ontology.Element result = null;
		Ontology.Element mumble = null;
		for (Enumeration o = children() ; o.hasMoreElements() ;) {
			mumble =  (Ontology.Element) o.nextElement();
			if (id.equalsIgnoreCase(mumble.id())) {
				result = mumble;
				break;
			} // end if
		} // end for
		if (result == null) {
			for (Enumeration o = children() ; o.hasMoreElements() ;) {
				mumble =  (Ontology.Element) o.nextElement();
				if (mumble instanceof Ontology.OntInterface) {
					result = ((Ontology.OntInterface) mumble).findElement(id);
					if (result != null) break;
				} // end if
			} // end for
		} // method
		return(result);
	} // method

	/** Finds the ontology with the specified id.
	*   Otherwise returns null.
	* * This could be made more efficient, especially with a different data
	*   structure, but how many ontologies are there going to be at any one
	*   time?  Probably less than a dozen.
	*/
	public static Ontology.OntInterface find (String id) {
		Ontology.OntInterface result = null;
		Ontology.OntInterface mumble = null;
		Ontology.find(id);
		return(result);
	} // method

	/** Returns true if the ontological element with the specified id
	*   exists in the ontology, otherwise returns false.
	* * Uses findElement and so recurses into categories.
	* @see #findElement(String)
	*/
	public boolean contains (String id) {
		return(findElement(id) != null);
	} // method

	/** Removes an ontological element from the ontology. */
	public void remove (Ontology.Element a) {
		elements().removeElement(a);
		a.category(null);
	} // method

	public void add (Instance instance) {
		super.add(instance);
		instances().removeElement(instance);
		instances().addElement(instance);
	} // method

	/** Adds an ontological element to the ontology. */
	public void add (Ontology.Element a) {
		if (a.category() != null) {
			Ontology.OntInterface mumble = find(a.category());
			if ((mumble != null) && (mumble instanceof Ontology))
				((Ontology) mumble).remove(a);
		} // end if
		elements().addElement(a);
		a.category(id());
	} // method

	/** Ontological elements contained in the ontology.
	* * Use accessors.
	* @see #elements()
	* @see #elements(Vector)
	* @serial
	*/
	protected Vector elements = null;
	/** Ontological elements contained in the ontology.
	* * Default is new Vector().
	* * Vector is very inefficient. Consider Linked List?
	*/
	public Vector elements () {
		if (elements == null) elements = new Vector();
		return(elements);
	} // reader
	/** @see #elements() */
	protected Vector elements (Vector v) { return(elements = v); }

	public Vector flatten () {
		Vector result = new Vector ();
		Object child = null;
		for(Enumeration kids = children(); kids.hasMoreElements();) {
			child = kids.nextElement();
			if(child instanceof Ontology) result = QRGutils.mergeVector(result, ((Ontology) child).flatten());
			else result.addElement(child);
		} // end for
		return(result);
	} // method

	public void remove (Instance instance) {
		super.remove(instance);
		instances.removeElement(instance);
	} // method

	public void clearProblems2 () {
		set("PROBLEMS", new Vector());
	}

	public Vector problems2 () { return(getVector("PROBLEMS")); }
	public Vector problems2 (String err, String msg) {
		Vector result = problems2();
		if (! result.contains(err)) {
			result.addElement(err);
			result.addElement(msg);
		} // end if
		return(result);
	} // method


	protected transient Hashtable bugs = null;
	public Hashtable bugs () {
		if (bugs == null) {
			bugs = new Hashtable();
		} // end if
		return(bugs);
	} // reader
	public void clearbugs () {
		bugs = null;
	} // reader
	public void clearbugs (Instance instance) {
		bugs().remove(instance.id());
	} // reader
	public void clearbugs (Model model) {
		bugs().remove(model.id());
	} // reader
	public void setbugs (Instance instance) {
		Vector value = new Vector();
		value.addElement(instance);
		value.addElement(instance.problems());
		bugs().put(instance.id(), value);
	} // reader
	public void setbugs (Model model) {
		Vector value = new Vector();
		value.addElement(model);
		value.addElement(model.problems2());
		bugs().put(model.id(), value);
	} // reader
	public Vector getbugs (String id) {
		Vector value = (Vector) bugs().get(id);
		Vector result = null;
		if (value != null) result = (Vector) value.elementAt(1);
		if (result == null) result = new Vector();
		return(result);
	} // reader
	public Vector getbugs (Instance instance) {
		Vector value = (Vector) bugs().get(instance.id());
		Vector result = null;
		if (value != null) result = (Vector) value.elementAt(1);
		if (result == null) result = new Vector();
		return(result);
	} // reader
	public Vector getbugs (Model model) {
		Vector value = (Vector) bugs().get(model.id());
		Vector result = null;
		if (value != null) result = (Vector) value.elementAt(1);
		if (result == null) result = new Vector();
		return(result);
	} // reader

	// NEED PREDICTION EXISTS CHECK & PREDICTION MET CHECK
	public static long checkPower = 0;
	public static final long CHECKNONE = 0;
	public static final long CHECKTARGET = (long) Math.floor(Math.pow(2.0, checkPower++));
	public static final long CHECKSYNTAX = (long) Math.floor(Math.pow(2.0, checkPower++));
	public static final long CHECKPROCESS = (long) Math.floor(Math.pow(2.0, checkPower++));
	public static final long CHECKINFLUENCE = (long) Math.floor(Math.pow(2.0, checkPower++));
	public static final long CHECKPREDICTIONEXISTS = (long) Math.floor(Math.pow(2.0, checkPower++));
	public static final long CHECKQLOOP = (long) Math.floor(Math.pow(2.0, checkPower++));
	public static final long CHECKABIGUITY = (long) Math.floor(Math.pow(2.0, checkPower++));
	public static final long CHECKPREDICTIONSVALID = (long) Math.floor(Math.pow(2.0, checkPower++));
	public static final long CHECKSTATE = (long) Math.floor(Math.pow(2.0, checkPower++));
	public static final long CHECKDESIGN = /* CHECKTARGET | */ CHECKSYNTAX | /* CHECKPROCESS | CHECKINFLUENCE | CHECKPREDICTIONEXISTS | */ CHECKQLOOP /* | CHECKABIGUITY */;
	public static final long CHECKGRAMMAR = CHECKPROCESS | CHECKINFLUENCE | CHECKPREDICTIONEXISTS | CHECKQLOOP | CHECKABIGUITY;
	public static final long CHECKALL = CHECKTARGET | CHECKSYNTAX | CHECKPROCESS | CHECKINFLUENCE | CHECKPREDICTIONEXISTS | CHECKQLOOP | CHECKABIGUITY | CHECKPREDICTIONSVALID | CHECKGRAMMAR | CHECKPREDICTIONSVALID | CHECKSTATE;

	public boolean checkModel (long checks, boolean clear) {
		boolean result = true;
		if (clear) clearProblems();
		if (result && ((checks & CHECKTARGET) != 0)) result = targetCheck();
		if (result && ((checks & CHECKSYNTAX) != 0)) result = syntaxCheck();
		if (result && ((checks & CHECKPROCESS) != 0)) result = processCheck();
		if (result && ((checks & CHECKINFLUENCE) != 0)) result = influenceCheck();
		if (result && ((checks & CHECKPREDICTIONEXISTS) != 0)) result = predictionExists();
		if (result && ((checks & CHECKQLOOP) != 0)) result = ! hasQpropLoop();
		if (result && ((checks & CHECKABIGUITY) != 0)) result = ambiguityCheck();
		if (result && ((checks & CHECKSTATE) != 0)) result = checkState();
		if (result && ((checks & CHECKPREDICTIONSVALID) != 0)) result = checkPredictions();
		return(result);
	} // method

	public String target () { return(getString("PROBLEM_STATEMENT")); }
	public String target (String s) { return((String) set("PROBLEM_STATEMENT", s)); }

	public boolean targetCheck () {
		boolean result = (! target().equals(""));
		if (! result) {
			problems2("Error #14", "You must specify a problem statement in your model target.");
			setbugs(this);
		} // end if
		return(result);
	} // method

	public boolean syntaxCheck () {
		boolean result = true;
		Instance instance = null;
		for(Enumeration instances = instances().elements(); instances.hasMoreElements();) {
			instance = (Instance) instances.nextElement();
			result = instance.checkInstance() && result;
		} // end for
		return(result);
	} // method

	public boolean processCheck () {
		boolean result = false;
		Instance instance = null;
		for(Enumeration instances = instances().elements(); instances.hasMoreElements();) {
			instance = (Instance) instances.nextElement();
			result = result || instance.typep(PROCESS);
			if (result) break;
		} // end for
		if (! result) {
			problems2("Error #15", "You must have a Process in your model.");
			setbugs(this);
		} // end if
		return(result);
	} // method

	public boolean influenceCheck () {
		boolean result = false;
		Instance instance = null;
		for(Enumeration instances = instances().elements(); instances.hasMoreElements();) {
			instance = (Instance) instances.nextElement();
			result = result || instance.typep(DIRECTINFLUENCE);
			if (result) break;
		} // end for
		if (! result) {
			problems2("Error #16", "You must have an Increases or Decreases in your model.");
			setbugs(this);
		} // end if
		return(result);
	} // method

	public boolean predictionExists () {
		boolean result = false;
		Instance instance = null;
		for(Enumeration instances = instances().elements(); instances.hasMoreElements();) {
			instance = (Instance) instances.nextElement();
			result = instance.typep( QUANTITY ) && (!instance.predictedState().equalsIgnoreCase( UNKNOWN ));
			if (result) break;
		} // end for
		if (! result) {
			problems2("Error #17", "You must make predictions about the behavior of your model.");
			setbugs(this);
		} // end if
		return(result);
	} // method

	public boolean checkPredictions () {
		boolean result = true;
		Instance instance = null;
		for(Enumeration instances = instances().elements(); instances.hasMoreElements();) {
			instance = (Instance) instances.nextElement();
			if (instance.typep( QUANTITY )) result = instance.checkPrediction() && result;
		} // end for
		return(result);
	} // method

	public boolean checkState () {
		boolean result = true;
		Instance instance = null;
		for(Enumeration instances = instances().elements(); instances.hasMoreElements();) {
			instance = (Instance) instances.nextElement();
			if (instance.typep(QUANTITY) && instance.state().equalsIgnoreCase(CONTRADICTION)) {
				result = false;
				instance.problems("Error #20", "This " + instance.name() + " as ambiguous inputs. Please assume a value or weight the input relationships. (Right-click.)");
			} // end if
		} // end for
		return(result);
	} // method

	public boolean ambiguityCheck () {
		boolean result = true;
		Vector increases = null;
		Vector decreases = null;
		Vector dominant = null;
		Vector normal = null;
		Vector negligable = null;
		Vector causes = null;
		Instance instance = null;
		String connectionID = null;
		Instance inputInstance = null;
		boolean check = false;
		for(Enumeration items = instances().elements(); items.hasMoreElements();) {
			instance = (Instance) items.nextElement();
			if (instance.typep(QUANTITY) || instance.typep(PROCESSRATE)) {
				increases = new Vector();
				decreases = new Vector();
				dominant = new Vector();
				normal = new Vector();
				negligable = new Vector();
				for(Enumeration connections = instance.causalInputs().elements(); connections.hasMoreElements();) {
					connectionID = (String) connections.nextElement();
					if (connectionID != null) {
						inputInstance = findInstance(connectionID);
						if (inputInstance != null) {
							if (inputInstance.state().equalsIgnoreCase(DOMINANT)) dominant.addElement(inputInstance);
							else if (inputInstance.state().equalsIgnoreCase(NEGLIGABLE)) negligable.addElement(inputInstance);
							else normal.addElement(inputInstance);
						} // end if
					} // end if
				} // end for
				if ( dominant.isEmpty() ) causes = normal;
				else causes = dominant;
				for(Enumeration connections = causes.elements(); connections.hasMoreElements();) {
					inputInstance = (Instance) connections.nextElement();
					if (inputInstance != null) {
						if (inputInstance.typep(IPLUS)) increases.addElement(inputInstance);
						if (inputInstance.typep(IMINUS)) decreases.addElement(inputInstance);
					} // end if
				} // end for
				check = (! ((increases.size() > 0) && (decreases.size() > 0)));
				if (! check) {
					instance.problems("Error #13.5", instance.name() + " is ambiguous.  It has both an increases and decreases as input at the same time. To fix, you can change the strength of the relationships or make an assumption about the parameter. (via Right-Click)"); // Decreases and Increases connected to same attribute
					result = false;
				} // end if
			} // end if
		} // end for
//		return(result);
		return(true);
	} // method

	public void updatePeers () {
		Instance instance = null;
		for(Enumeration instances = instances().elements(); instances.hasMoreElements();) {
			instance = (Instance) instances.nextElement();
			instance.updatePeer();
		} // end for
	} // method

	/**
	 * Have to break apart the setting and resetting of problems in order to handle
	 * global problems.
	 **/
	void clearProblems () {
		clearbugs();
		clearProblems2();
		for(Enumeration instances = instances().elements(); instances.hasMoreElements();) {
			Instance instance = (Instance) instances.nextElement();
			instance.clearProblems();
		}
	}

	public int mark = 0;			// used in marker algorithms for loop detection
	private int startmark = 0;		// used in marker algorithms for loop detection

	/**
	 * Iterate over all QPROP relations and for each one, check if it is in a loop
	 **/
	boolean hasQpropLoop()
	{
		boolean result = false;
		mark++;
		startmark = mark;
		for ( Enumeration items = connections().elements(); items.hasMoreElements(); )
		{
			Instance relationship = (Instance)items.nextElement();
			if (relationship.typep(INDIRECTINFLUENCE))
			{
				if (inQpropLoop(relationship) == true)
				{
					if (relationship.mark == mark + 1)
						mark++;
					markErrors();
					result = true;
					break;
				}
				else
					mark++;
			}
		}
//		set("CHECKQLOOPS", (result));
		return(result);
	}

	/**
	 * Determine whether a particular quantity is in a QProp loop.
	 **/
	boolean inQpropLoop(Instance rel)
	{
		//QRGutils.debug("Calling inQpropLoop on " + describeRel(rel));
		if (rel.mark >= startmark) // been here before?
			return false;
		rel.mark = mark;			// mark this node.
		if (rel.input() == null || rel.input() == "") return false;	// dead-end
		for (Enumeration influences = connections().elements(); influences.hasMoreElements(); )
		{
			Instance r = (Instance)influences.nextElement();
			if (rel.input() != r.output()) continue;
			if (!r.typep(INDIRECTINFLUENCE)) continue;
			if (r.mark == mark)
			{

				//QRGutils.debug("found cycle at: " + describeRel(r));
				r.mark = mark + 1;
				rel.mark = mark + 1;
				return true;
			}
			else if (inQpropLoop(r))
			{
				// Update the marks as we unwind the stack in order to just
				// label those nodes in the cycle.
				if (rel.mark == mark + 1)
					mark++;	// we've unwound the stack to the start of the cycle
				else if (rel.mark == mark)
					rel.mark = mark + 1;
				return true;
			}
		}
		return false;
	}

	// For debugging:
	String describeRel(Instance r)
	{
		Instance i = findInstance(r.input());
		Instance o = findInstance(r.output());
		String result = r.name() + " from " + i.name() + " to " + o.name();
		return result;
	}

	/**
	 * For each relationship with the current mark, add a problem description to
	 * the problems vector of the relationship and its inputs and outputs.
	 **/
	void markErrors()
	{
		for (Enumeration influences = connections().elements(); influences.hasMoreElements(); )
		{
			Instance r = (Instance)influences.nextElement();
			if (r.typep(QPROPPLUS) ||  r.typep(QPROPMINUS))
			{
				if (r.mark == mark)
				{
					r.problems("Error #??", "This is part of a loop of influences");
					setbugs(r);
					Instance output = findInstance(r.output());
					if (output != null) // shouldn't happen
					{
						output.problems("Error #???", "This is part of a loop of influences");
						setbugs(output);
					}
				}
			}
		}
	}


} // class Model

// ---------------------------------------------------------------------------
// End of File: Model.java                    (c) 2003 Northwestern University
